"""Model Commands."""
import fnmatch
from dataclasses import dataclass
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import typer
import validators  # type: ignore
from huggingface_hub import snapshot_download
from rich.markdown import Markdown
from toolz import itertoolz
from toolz import recipes

from flyswot.config import APP_NAME
from flyswot.config import MODEL_REPO_ID
from flyswot.console import console

app = typer.Typer()


@dataclass
class LocalModel:
    """A local model container"""

    hf_cache: Path

    def __post_init__(self):
        """Returns model parts contained under hf_cache"""
        self._get_model_parts(self.hf_cache)

    def _get_model_parts(self, hf_cache: Path):
        """Returns model path, vocab and metadata for a model"""
        model_files = Path(hf_cache).iterdir()
        for file in model_files:
            if fnmatch.fnmatch(file.name, "vocab.txt"):
                self.vocab = file
            if fnmatch.fnmatch(file.name, "README.md"):
                self.modelcard = file
            if fnmatch.fnmatch(file.name, "*.onnx") or fnmatch.fnmatch(
                file.name, "*.pkl"
            ):
                self.model = file


def _url_callback(url: str) -> Union[str, None]:
    """Checks url is valid"""
    if url == "latest":
        return url
    if validators.url(url):
        return url
    else:
        raise typer.BadParameter(f"Please check {url} is a valid url")


def ensure_model_dir(model_dir_path: Union[Path, None] = None) -> Path:
    """Checks for a local model dir and creates one if not found"""
    if not model_dir_path:
        app_dir = typer.get_app_dir(APP_NAME)
        model_dir: Path = Path(app_dir) / "models"
    else:
        model_dir = Path(model_dir_path) / "models"
    if not (model_dir.exists() and model_dir.is_dir()):
        typer.echo(f"Creating directory for storing models in {model_dir}...")
        try:
            model_dir.mkdir(parents=True)
        except PermissionError as e:  # pragma: no cover
            typer.echo(f"{model_dir} is not writeable: {e}")
            raise typer.Exit(code=1) from None
    typer.echo(f"Models stored in {model_dir}")
    return model_dir


@app.command()
def show_model_dir() -> None:
    """Print out the directory where models are stored"""
    ensure_model_dir()


@app.command(name="download")
def get_model(
    revision: Optional[str] = typer.Argument(None, callback=_url_callback),
    model_dir: Path = typer.Argument(
        None,
        envvar="MODEL_DIR",
        help="Optionally specify a directory to store model files in",
    ),
    local_only=False,
) -> Path:  # pragma: no cover
    """Downloads models, defaults to the latest available model"""
    repo_id = MODEL_REPO_ID
    with console.status("Getting model", spinner="dots"):
        model = snapshot_download(
            repo_id, cache_dir=model_dir, revision=None, local_files_only=local_only
        )
    return Path(model)


# def _get_model_parts(model_dir: Path) -> Optional[LocalModel]:
#     """Returns model path, vocab and metadata for a model"""
#     model_files = Path(model_dir).iterdir()
#     vocab = None
#     modelcard = None
#     model = None
#     for file in model_files:
#         if fnmatch.fnmatch(file.name, "vocab.txt"):
#             vocab = file
#         if fnmatch.fnmatch(file.name, "README.md"):
#             modelcard = file
#         if fnmatch.fnmatch(file.name, "*.onnx") or fnmatch.fnmatch(file.name, "*.pkl"):
#             model = file
#     if vocab and modelcard and model is not None:
#         return LocalModel(vocab, modelcard, model)


# def load_model(model_dir: Path):  # pragma: no cover
#     """returns a local model"""
#     return _get_model_parts(model_dir)


def ensure_model(model_dir: Path) -> LocalModel:  # pragma: no cover
    """Checks for a local model and if not found downloads the latest available remote model"""
    model = get_model(model_dir=model_dir)
    if model:
        return LocalModel(model)
    typer.echo("Not able to find a model")
    raise typer.Exit()


def is_pipe(c: Tuple) -> bool:
    """Checks if | in c"""
    return "|" in c


def load_vocab(vocab: Path) -> List[List[str]]:
    """loads vocab from `vocab` and returns as list contaning lists of vocab"""
    with open(vocab, "r") as f:
        raw_vocab = [line.strip("\n") for line in f.readlines()]
        return list(
            map(
                list,
                (itertoolz.remove(is_pipe, (recipes.partitionby(is_pipe, raw_vocab)))),
            )
        )


@app.command()
def vocab(model: str = typer.Argument("latest"), show: bool = typer.Option(True)):
    """Prints out vocab for latest model"""
    if model != "latest":
        raise NotImplementedError
    model_dir = ensure_model_dir()
    model_path = get_model(model_dir=model_dir)
    if model_path:  # pragma: no cover
        local_model = LocalModel(model_path)
        if local_model.vocab:
            vocab = load_vocab(local_model.vocab)
            if show:
                console.print(Markdown("# Model Vocab"))
                console.print(vocab)
            return vocab


def show_model_card(localmodel: LocalModel):
    """Shows model card for model"""
    with open(localmodel.modelcard, "r") as f:
        md = Markdown(f.read())
    console.print(md)


if __name__ == "__main__":  # pragma: no cover
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['flyswot']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8,<10',
 'huggingface-hub>=0.2.1,<0.5.0',
 'numpy>=1.20.2,<2.0.0',
 'onnxruntime>=1.7.0,<2.0.0',
 'rich>=10.1,<12.0',
 'toolz>=0.11.1,<0.12.0',
 'typer>=0.3.2,<0.5.0',
 'typing_extensions>=3.10,<5.0',
 'validators>=0.18.2,<0.19.0']

entry_points = \
{'console_scripts': ['flyswot = flyswot.cli:main']}

setup_kwargs = {
    'name': 'flyswot',
    'version': '0.3.1',
    'description': 'flyswot',
    'long_description': 'flyswot\n=======\n\n|PyPI| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/flyswot.svg\n   :target: https://pypi.org/project/flyswot/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/flyswot\n   :target: https://pypi.org/project/flyswot\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/flyswot\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/flyswot/latest.svg?label=Read%20the%20Docs\n   :target: https://flyswot.readthedocs.io/\n   :alt: Read the documentation at https://flyswot.readthedocs.io/\n.. |Tests| image:: https://github.com/davanstrien/flyswot/workflows/Tests/badge.svg\n   :target: https://github.com/davanstrien/flyswot/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/davanstrien/flyswot/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/davanstrien/flyswot\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n.. image:: https://raw.githubusercontent.com/davanstrien/flyswot/main/docs/_static/fly.png?token=ACEUI5KJ2HPO4ZGNTBX6OE3ARMXII\n\n\nDisclaimer\n-----------\n\n`flyswot` is a work in progress and is currently only intended to be used for testing by `HMD <https://www.bl.uk/projects/heritage-made-digital>`_\n\n\nThis code and documentation is a work in progress.\n\nFeatures\n--------\n\n`flyswot` is a Command Line Tool for detecting \'fake\' flysheets.\n\n* unix style search patterns for matching images to predict against\n* produces a csv output containing the paths to the input images, the predicted label and the models confidence for that prediction.\n* produces a summary \'report\' providing a high level summary of the predictions made by `flyswot`\n* automatically downloads latest available `flyswot model <https://huggingface.co/davanstrien/flyswot>`_\n\n.. image:: https://asciinema.org/a/449685.svg\n   :target: https://asciinema.org/a/449685\n\n\nRequirements\n------------\n\n* Python 3.7 or greater\n\nTODO update this section with requirements and uses.\n\nInstallation\n------------\n\nYou can install *flyswot* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install flyswot\n\nThis will install the latest release version of *flyswot*\n\nDetailed Installation Guide\n---------------------------\n\nThis section gives a more detailed instructions for installing *flyswot*. This guidance is aimed particularly at `HMD`_ users of *flyswot*. This set of instructions covers the steps required to install *flyswot*.\n\nNote on the examples\n^^^^^^^^^^^^^^^^^^^^\n\nYou will see examples for input in the guidance below which looks like:\n\n.. code:: console\n\n          $ pip\n\nThe ``$`` symbol here is often used as a convention to show that this is input to a terminal/command line. When you input this into your own terminal you should only input the part after the ``$`` symbol. For example, in the above case you would type ``pip``.\n\n\nInstall Python\n^^^^^^^^^^^^^^\n\n*flyswot* uses the `Python`_ programming language. You will therefore need to have Python installed on your computer to run *flyswot*.\n\nFor HMD users of `flyswot` it is suggested to use the `Anaconda`_ distribution of Python. If you are on a managed PC/laptop you should request this to be installed via the Technology team.\n\nCreate and activate a Conda Environment\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nA virtual environment allows us to isolate the requirements of different python packages. This can be useful since different python packages might have different requirements. Using a virtual environment allows us to install these in a way where you are less likely to have conflicts between these packages.\n\nThere are various different ways of creating virtual environments within Python. Anaconda comes with a system for creating virtual environment\'s. Creating Conda environments within Conda can be done in various ways, if you are new to Python and Anaconda it is suggested to use `Anaconda Navigator`_.\n\nYou can find more detailed instructions for creating a new virtual environments in the `Anaconda documentation`_. Briefly you should:\n\n- Open Anaconda Navigator\n- Navigate to the *Environments* tab\n- Create a new Python environment and give it a descriptive name i.e. *flyswot*. You should ensure that the Python version is version 3.7 or above.\n\nOnce you have created this new environment you can "activate it" by clicking on the arrow next to the name of the environment. You should select the "open terminal". This should open a new terminal window. On the left you should see the name of your environment in brackets:\n\n.. code:: console\n\n    (flyswot) $\n\nInstall flyswot in your Conda Environment\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nNow you have created and activated your conda environment you can install flyswot. Before doing this you can check that `pip`_ is available inside your environment:\n\n.. code:: console\n\n          $ pip\n\nThis should return the help pages for `pip`_. We can now use `pip`_ to install *flyswot* from `PyPI`_.\nTo do this run:\n\n.. code:: console\n\n   $ pip install flyswot\n\nYou should now see the installation process begin. During this you will see some information printed to the terminal abou what packages are being installed.\n\nCheck flyswot is installed\n^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nWe can quickly confirm that *flyswot* has been installed by printing out the help information. We can do this by typing:\n\n.. code:: console\n\n  $ flyswot --help\n\nThis should print some help information to the screen. If you get this it means you have successfully installed *flyswot* 💪.\n\nUpdate flyswot\n^^^^^^^^^^^^^^^\n\nIf you need to upgrade your version of flyswot you should activate your Conda Environment and then run\n\n.. code:: console\n\n   $ pip install flyswot --upgrade\n\n\nUsage\n-----\n\nTo run predictions against a directory of images:\n\n.. code:: console\n\n   $ flyswot predict directory manuscripts_folder .\n\n- *flyswot* will search inside the manuscripts_folder looking for image files.\n- By default it will look for files that contain :code:`FS` in the filename since these are files which have been labelled as being "end flysheets" or "front flysheets"\n- Once it has found all the files labelled as `flysheet` it will then run a computer vision model against these images to see if they are labelled correctly i.e. if it is indeed a flysheet or something else.\n- flyswot will save a csv report containing the paths to the image, the directory the image is stored in, the label, and the confidence for that prediction.\n\nDetailed Usage Guide\n--------------------\n\nThis section provides additional guidance on the usage of *flyswot*. This is primarily aimed at `HMD`_ users of *flyswot*.\n\nHow flyswot searches for images\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n*flyswot* is currently intended to identify images which have an incorrect label associated with them. In particular it is currently intended to identify "fake" flysheets. These images have ``fs`` as part of their filename so this is used by *flyswot* to identify images which should be checked using the computer vision model. This can be channged if you also want to match other filename patterns.\n\nSince these images of concern will often be inside a directory structure *flyswot* will look in sub-folders from the input folder for images which contain ``fs`` in the name. For example in the following folder structure:\n\n.. code:: console\n\n   Collection/\n   ├─ item1/\n   │  ├─ add_ms_9403_fbspi.tif\n   │  ├─ add_ms_9403_fse001r.tif\n   │  ├─ add_ms_9403_fse001v.tif\n   ├─ item2/\n   │  ├─ sloane_ms_116_fblefr.tif\n   │  ├─ sloane_ms_116_fbspi.tif\n   │  ├─ sloane_ms_116_fse004r.tif\n\nAll of the files which have ``fs`` in the filname will be check but files which don\'t contains `fs` such as ``add_ms_9403_fbspi.tif`` will be ignored since these aren\'t labelled as flysheets.\n\nRunning flyswot against a directory of images\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nTo run *flyswot* against a directory of images you need to give it the path to that directory/folder.\nThere are different ways you could do this. The following is suggested for people who are not very familiar (yet 😜) with terminal interfaces.\n\nIdentify the folder you want to *flyswot* to check for "fake" flysheets. If you are using *flyswot* for the first time it may make sense to choose a folder which doesn\'t contain a huge number of collection items so you don\'t have to wait to long for *flyswot* to finish running. Once you have found a directory you want to predict against copy the path. This path should be the full path to the item.\n\nFor example something that looks like:\n\n.. code:: console\n\n  \\\\ad\\collections\\hmd\\excitingcollection\\excitingsubcollection\\\n\n\nThis will be the folder from which *flyswot* starts looking.\n\nWhen you activated your conda environment in a terminal, you were likely \'inside\' your user directory. Since we need to specify a place for *flyswot* to store the CSV report, we\'ll move to a better place to store that output; your ``Desktop`` folder. To do we can navigate using the command:\n\n.. code:: console\n\n   $ chdir desktop\n\n\nif you are using  Mac, Linux or have GitBash installed you should instead run:\n\n.. code:: console\n\n   $ cd Desktop\n\n\n\nThis will take you to your Desktop. We\'ll now run *flyswot*. As with many other command line tools, *flyswot* has commands and sub-commands. We are interested in the ``predict`` command. This includes two sub-commands: ``predict-image`` and ``directory``.  We will mostly want to predict directories. To do this we use the following approach\n\n.. code:: console\n\n   $ flyswot predict directory input_directory output_directory\n\nThe input directory is the folder containing our images and the output directory is where we want to save our CSV report. Using the folder we previously identified this would look like:\n\n.. code:: console\n\n   $ flyswot predict directory "\\\\ad\\collections\\hmd\\excitingcollection\\excitingsubcollection\\" .\n\nWe can use ``.`` to indicate we want the CSV report to be saved to the current directory (in this case the Deskop directory). Also notice that there are quotation marks ``""`` around the path. This is used to make sure that any spaces in the path are escaped.\n\nOnce you run this command you should see some progress reported by *flyswot*, including a progress bar that shows how many of the images *flyswot* has predicted against.\n\nWhen *flyswot* has finshed you will have a CSV \'report\' which contains the path to the image, the predicted label and the confidence for that prediction.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*flyswot* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/davanstrien/flyswot/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://flyswot.readthedocs.io/en/latest/usage.html\n.. _HMD: https://www.bl.uk/projects/heritage-made-digital\n.. _Python: https://www.python.org/\n.. _Anaconda: https://www.anaconda.com/products/individual\n.. _Anaconda Navigator: https://docs.anaconda.com/anaconda/navigator/\n.. _Anaconda Documentation: https://docs.anaconda.com/anaconda/navigator/tutorials/manage-environments/#creating-a-new-environment\n',
    'author': 'Daniel van Strien',
    'author_email': 'daniel.van-strien@bl.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/davanstrien/flyswot',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



"""Inference functionality"""
import csv
import mimetypes
import re
import string
import time
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from collections import OrderedDict
from dataclasses import asdict
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from functools import singledispatch
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import onnxruntime as rt
import PIL
import typer
from PIL import Image  # type: ignore
from rich import print
from rich.columns import Columns
from rich.layout import Layout
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from toolz import itertoolz

from flyswot import core
from flyswot import models
from flyswot.console import console
from flyswot.logo import flyswot_logo

app = typer.Typer()


@dataclass
class ImagePredictionItem:
    """Prediction for an image.

    Attributes:
        path: The Path to the image
        predicted_label: The predicted label i.e. the argmax value for the prediction tensor
        confidence: The confidence for `predicted_label` i.e. the max value for prediction tensor
    """

    path: Path
    predicted_label: str
    confidence: float

    def __post_init__(self) -> Union[Path, None]:
        """attempt to get absolute path"""
        try:
            self.path: Path = self.path.absolute()
        except AttributeError:
            pass


@dataclass
class MultiLabelImagePredictionItem:
    """Multiple predictions for a single image"""

    path: Path
    predictions: List[Dict[float, str]]

    def _get_top_labels(self) -> List[str]:
        """Get top labels"""
        top_labels = []
        for prediction in self.predictions:
            top_label = sorted(prediction.items(), reverse=True)[0][1]
            top_labels.append(top_label)
        return top_labels

    # Post init that gets top prediction label from predictions
    def __post_init__(self):
        """Get top prediction label"""
        self.predicted_labels = self._get_top_labels()


@dataclass
class PredictionBatch:
    """Container for ImagePredictionItems"""

    batch: List[ImagePredictionItem]

    def __post_init__(self):
        """Returns a iterable of all predicted labels in batch"""
        self.batch_labels: Iterable[str] = (item.predicted_label for item in self.batch)


@dataclass
class MultiPredictionBatch:
    """Container for MultiLabelImagePredictionItems"""

    batch: List[MultiLabelImagePredictionItem]

    def _get_predicted_labels(self) -> Iterable:
        """Returns a iterable of all predicted labels in batch"""
        return (item.predicted_labels for item in self.batch)

    def __post_init__(self):
        """Returns a iterable of lists containing all predicted labels in batch"""
        # self.batch_labels: Iterable = (
        #     list(itertoolz.pluck(0, pred))
        #     for pred in zip(*[o.predictions for o in self.batch])
        # )
        self.batch_labels = self._get_predicted_labels()


image_extensions = {k for k, v in mimetypes.types_map.items() if v.startswith("image/")}


@app.command()
def predict_image(
    image: Path = typer.Argument(..., readable=True, resolve_path=True)
) -> None:
    """Predict a single image"""
    pass  # pragma: no cover


def check_files(files: List, pattern: str, directory: Path) -> None:
    """Check if files exist and raises error if not"""
    if not files:
        typer.echo(
            f"Didn't find any files maching {pattern} in {directory}, please check the inputs to flyswot"
        )
        raise typer.Exit(code=1)


def try_predict_batch(batch, inference_session, bs):
    """try and predict a batch of files"""
    bad_batch = False
    try:
        batch_predictions = inference_session.predict_batch(batch, bs)
        return batch_predictions, bad_batch
    except PIL.UnidentifiedImageError:
        typer.echo("Corrupt imagess found in batch")
        bad_batch = True
        return batch, bad_batch


def predict_files(files: List[Path], inference_session, bs, csv_fname) -> set:
    """Predict files"""
    with typer.progressbar(length=len(files)) as progress:
        images_checked = 0
        bad_batch_files = []
        for i, batch in enumerate(itertoolz.partition_all(bs, files)):
            batch_predictions, bad_batch = try_predict_batch(
                batch, inference_session, bs
            )
            if bad_batch:
                bad_batch_files.append(batch)
            if i == 0 and not bad_batch:
                create_csv_header(batch_predictions, csv_fname)
            if not bad_batch:
                write_batch_preds_to_csv(batch_predictions, csv_fname)
            progress.update(len(batch))
            images_checked += len(batch)
        corrupt_images = set()
        for batch in bad_batch_files:
            for file in batch:
                try:
                    batch_predictions = inference_session.predict_batch(files, bs)
                    write_batch_preds_to_csv(batch_predictions)
                except PIL.UnidentifiedImageError:
                    corrupt_images.add(file)
        return corrupt_images, images_checked


@app.command(name="directory")
def predict_directory(
    directory: Path = typer.Argument(
        ...,
        readable=True,
        resolve_path=True,
        help="Directory to start searching for images from",
    ),
    csv_save_dir: Path = typer.Argument(
        ...,
        writable=True,
        resolve_path=True,
        help="Directory used to store the csv report",
    ),
    pattern: str = typer.Option("fs", help="Pattern used to filter image filenames"),
    bs: int = typer.Option(16, help="Batch Size"),
    image_format: str = typer.Option(
        ".tif",
        help="Image format for flyswot to use for predictions, defaults to `*.tif`",
    ),
    model_name: str = typer.Option(
        "latest", help="Which model flyswot should use for making predictions"
    ),
    model_path: str = None,
):
    """Predicts against all images stored under DIRECTORY which match PATTERN in the filename.

    By default searches for filenames containing 'fs'.

    Creates a CSV report saved to `csv_save_dir`
    """
    start_time = time.perf_counter()
    model_dir = models.ensure_model_dir()
    model = models.ensure_model(model_dir)
    onnxinference = OnnxInferenceSession(model.model, model.vocab)
    files = sorted(core.get_image_files_from_pattern(directory, pattern, image_format))
    check_files(files, pattern, directory)
    typer.echo(f"Found {len(files)} files matching {pattern} in {directory}")
    csv_fname = create_csv_fname(csv_save_dir)
    corrupt_images, images_checked = predict_files(
        files, inference_session=onnxinference, bs=bs, csv_fname=csv_fname
    )
    if corrupt_images:
        print(corrupt_images)
    delta = timedelta(seconds=time.perf_counter() - start_time)
    print_inference_summary(
        str(delta), pattern, directory, csv_fname, image_format, images_checked
    )


def print_inference_summary(
    time_delta: str,
    pattern: str,
    directory: Path,
    csv_fname: Path,
    image_format: str,
    matched_file_count: int,
    local_model: Optional[models.LocalModel] = None,
):
    """prints summary report"""
    print(flyswot_logo())
    print(
        Panel(
            Text(
                f"CSV report file at: {csv_fname.as_uri()}",
            ),
            title=" :clipboard: CSV report :clipboard:",
        )
    )
    print(
        Panel(
            Text(f"{time_delta}", justify="center", style="bold green"),
            title=":stopwatch: Time taken to run :stopwatch:",
        )
    )
    print(
        create_file_summary_markdown(
            pattern, matched_file_count, directory, image_format
        )
    )
    inference_summary_columns = get_inference_table_columns(csv_fname)
    print(Panel(inference_summary_columns, title="Prediction Summary"))
    if local_model is not None:
        print(models.show_model_card(localmodel=local_model))


def create_file_summary_markdown(
    pattern: str, matched_file_count: int, directory: Path, image_format: str
) -> Panel:
    """creates Markdown summary containing number of files checked by flyswot vs total images files under directory"""
    total_image_file_count = core.count_files_with_ext(directory, image_format)
    return Panel(
        Markdown(
            f"""
    - flyswot searched for image files by matching the patern *{pattern}*
    - flyswot search inside: `{directory}`
    - In total the directory contained **{total_image_file_count}** images
    - There were **{matched_file_count}** files matching the {pattern}* pattern which flyswot checked
    """
        ),
        title=":file_folder: files checked :file_folder:",
    )


def get_inference_table_columns(csv_fname: Path) -> Columns:
    """print_inference_summary from `fname`"""
    labels_to_print = labels_from_csv(csv_fname)
    tables = [
        print_table(labels, f"Prediction summary {i+1}", print=False)
        for i, labels in enumerate(labels_to_print)
    ]
    return Columns(tables)


label_regex = re.compile(r"prediction_label_\D_0")


def labels_from_csv(fname: Path) -> List[List[str]]:
    """Gets top labels from csv `fname`"""
    columns = defaultdict(list)
    with open(fname, "r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            for (k, v) in row.items():
                columns[k].append(v)
    return [columns[k] for k in columns if label_regex.match(k)]


def print_table(
    decoded: list, header: str = "Prediction summary", print: bool = True
) -> Table:
    """Prints table summary of predicted labels"""
    table = Table(show_header=True, title=header)
    table.add_column(
        "Class",
    )
    table.add_column("Count")
    table.add_column("Percentage")
    total = len(decoded)
    frequencies = itertoolz.frequencies(decoded)
    for is_last_element, var in core.signal_last(frequencies.items()):
        key, value = var
        count = value
        percentage = round((count / total) * 100, 2)
        if is_last_element:
            table.add_row(key, str(count), f"{percentage}", end_section=True)
            table.add_row("Total", str(total), "")
        else:
            table.add_row(key, str(count), f"{percentage}")
    if print:
        console.print(table)
    return table


def make_layout():
    """Define the layout."""
    layout = Layout(name="root")
    layout.split(Layout(name="header", size=4), Layout(name="main"))
    layout["main"].split_column(
        Layout(name="info", size=4), Layout(name="body", ratio=2, minimum_size=60)
    )
    layout["info"].split_row(Layout(name="time"), Layout(name="files"))
    return layout


# def print_summary(columns, time, files):
#     layout = make_layout()
#     MARKDOWN = """
#     # Prediction summary
#     """
#     md = Markdown(MARKDOWN)
#     layout["header"].update(md)
#     layout["body"].update(columns)
#     layout["time"].update(time)
#     layout["files"].update(f"Number of files checked {len(files)}")
#     print(layout)


def create_csv_fname(csv_directory: Path) -> Path:
    """Creates a csv filename"""
    date_now = datetime.now()
    date_now = date_now.strftime("%Y_%m_%d_%H_%M")
    fname = Path(date_now + ".csv")
    return Path(csv_directory / fname)


@singledispatch
def create_csv_header(batch, csv_path) -> None:
    """Print csv header"""
    raise NotImplementedError(f"Not implemented for type {batch}")


@create_csv_header.register
def _(batch: PredictionBatch, csv_path):
    """Creates a header for csv `csv_path`"""
    with open(csv_path, mode="w", newline="") as csv_file:
        field_names = ["path", "directory", "prediction", "confidence"]
        writer = csv.DictWriter(csv_file, fieldnames=field_names)
        writer.writeheader()


@create_csv_header.register
def _(batch: MultiPredictionBatch, csv_path: Path, top_n: int = 2):
    item = batch.batch[0]
    pred = OrderedDict()
    pred["path"] = asdict(item)["path"]
    pred["directory"] = asdict(item)["path"].parent
    for i, _ in enumerate(item.predictions):
        for j in range(top_n):
            pred[f"prediction_label_{string.ascii_letters[i]}_{j}"] = ""
            pred[f"confidence_label_{string.ascii_letters[i]}_{j}"] = ""
    with open(csv_path, mode="w", newline="") as csv_file:
        field_names = list(pred.keys())
        writer = csv.DictWriter(csv_file, fieldnames=field_names)
        writer.writeheader()


@singledispatch
def write_batch_preds_to_csv(predictions, csv_fpath: Path) -> None:
    """print batch preds to csv"""
    raise NotImplementedError(f"Not implemented for type {predictions}")


@write_batch_preds_to_csv.register
def _(predictions: PredictionBatch, csv_fpath: Path) -> None:
    """Appends `predictions` batch to `csv_path`"""
    with open(csv_fpath, mode="a", newline="") as csv_file:
        field_names = ["path", "directory", "predicted_label", "confidence"]
        writer = csv.DictWriter(csv_file, fieldnames=field_names)
        for pred in predictions.batch:
            row = asdict(pred)
            row["directory"] = pred.path.parent
            writer.writerow(row)


@write_batch_preds_to_csv.register
def _(predictions: MultiPredictionBatch, csv_fpath: Path, top_n: int = 2) -> None:
    for pred in predictions.batch:
        row = OrderedDict()
        row["path"] = asdict(pred)["path"]
        row["directory"] = asdict(pred)["path"].parent
        for i, v in enumerate(pred.predictions):
            sorted_predictions = sorted(v.items(), reverse=True)
            for j in range(top_n):
                row[f"prediction_label_{i}_{j}"] = sorted_predictions[j][1]
                row[f"confidence_label_{i}_{j}"] = sorted_predictions[j][0]
        with open(csv_fpath, mode="a", newline="") as csv_file:
            field_names = list(row.keys())
            writer = csv.DictWriter(csv_file, fieldnames=field_names)
            writer.writerow(row)


class InferenceSession(ABC):
    """Abstract class for inference sessions"""

    @abstractmethod
    def __init__(self, model: Path, vocab: List):  # pragma: no cover
        """Inference Sessions should init from a model file and vocab"""
        self.model = model
        self.vocab = vocab

    @abstractmethod
    def predict_image(self, image: Path):  # pragma: no cover
        """Predict a single image"""
        pass

    @abstractmethod
    def predict_batch(
        self, model: Path, batch: Iterable[Path], bs: int
    ):  # pragma: no cover
        """Predict a batch"""
        pass


def softmax(x):
    """return softmax of `x`"""
    x = x.reshape(-1)
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)


# class FastaiInferenceModel(InferenceSession):
#     def __init__(self, model):
#         self.model = model
#         self.learn = load_learner(model)

#     def predict_image(self, image: Path) -> Any:
#         return self.learn.predict(image)

#     def predict_batch(self, batch: Iterable[Path], bs: int) -> PredictionBatch:
#         test_dl = self.learn.dls.test_dl(batch, bs=bs)
#         vocab = dict(enumerate(self.learn.dls.vocab))
#         with self.learn.no_bar():
#             fastai_preds: Any = self.learn.get_preds(dl=test_dl, with_decoded=True)
#             prediction_tensors: Iterable[Any] = fastai_preds[0]
#             prediction_items = []
#             for file, pred in zip(batch, prediction_tensors):
#                 arg_max = int(np.array(pred).argmax())
#                 predicted_label = vocab[int(arg_max)]
#                 confidence = float(np.array(pred).max())
#                 prediction_items.append(
#                     ImagePredictionItem(file, predicted_label, confidence)
#                 )
#         return PredictionBatch(prediction_items)


class OnnxInferenceSession(InferenceSession):  # pragma: no cover
    """onnx inference session"""

    def __init__(self, model: Path, vocab: Path):
        """Create onnx session"""
        self.model = model
        self.session = rt.InferenceSession(str(model))

        self.vocab = models.load_vocab(vocab)
        self.vocab_mappings = [dict(enumerate(v)) for v in self.vocab]

    def predict_image(
        self, image: Path
    ) -> Union[ImagePredictionItem, MultiLabelImagePredictionItem]:
        """Predict a single image"""
        img = self._load_image(image)
        output_names = [o.name for o in self.session.get_outputs()]
        raw_result = self.session.run(output_names, {"image": img})
        if len(self.vocab) < 2:
            pred = self._postprocess(raw_result)
            arg_max = int(np.array(pred).argmax())
            predicted_label = self.vocab_mappings[0][int(arg_max)]
            confidence = float(np.array(pred).max())
            return ImagePredictionItem(image, predicted_label, confidence)
        else:
            prediction_dicts = []
            for vocab_map, pred in zip(self.vocab_mappings, raw_result):
                predictions = self._postprocess(pred)
                prediction_dict = {
                    float(prediction): vocab_map[i]
                    for i, prediction in enumerate(predictions)
                }
                prediction_dicts.append(prediction_dict)
        return MultiLabelImagePredictionItem(image, prediction_dicts)

    def _preprocess(self, input_data: np.ndarray) -> np.ndarray:
        # converts the input data into the float32 input for onnx
        img_data = input_data.astype("float32")

        # normalize
        mean_vec = np.array([0.485, 0.456, 0.406])
        stddev_vec = np.array([0.229, 0.224, 0.225])
        norm_img_data = np.zeros(img_data.shape).astype("float32")
        for i in range(img_data.shape[0]):
            norm_img_data[i, :, :] = (
                img_data[i, :, :] / 255 - mean_vec[i]
            ) / stddev_vec[i]

        # add batch channel
        norm_img_data = norm_img_data.reshape(1, 3, 512, 512).astype("float32")
        return norm_img_data

    def _load_image(self, file: Path) -> np.ndarray:
        """loads image and carries out preprocessing for inference"""
        image = Image.open(file, mode="r")
        image = image.resize((512, 512), Image.BILINEAR)
        image_data = np.array(image).transpose(2, 0, 1)
        return self._preprocess(image_data)

    def _postprocess(self, result: List):
        """process results from onnx session"""
        return softmax(np.array(result)).tolist()

    def predict_batch(
        self, batch: Iterable[Path], bs: int
    ) -> Union[PredictionBatch, MultiPredictionBatch]:
        """predicts a batch of images"""
        prediction_items = [self.predict_image(file) for file in batch]
        if len(self.vocab) < 2:
            return PredictionBatch(prediction_items)
        else:
            return MultiPredictionBatch(prediction_items)


if __name__ == "__main__":
    app()  # pragma: no cover
